#! /bin/bash

# Copyright (C) 2007 Stephane Levant <stephane.levant@gmail.com>

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2, or (at your option)
#  any later version.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; see the file COPYING.  If not, write to the
#  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.

VERSION=0.2
prog=`basename $0`

### Locales

L_VERSION=(
'Copyright (C) 2007 Stephane Levant <stephane.levant@gmail.com>
This is free software.  You may redistribute copies of it under the terms of
the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.
There is NO WARRANTY, to the extent permitted by law.
'
'Copyright (C) 2007 Stephane Levant <stephane.levant@gmail.com>
Ce programme est un logiciel libre. Vous pouvez en redistribuer des copies
selon les termes de la License Publique Gnrale de GNU, 
<http://www.gnu.org/licenses/gpl.html>.
AUCUNE GARANTIE n'"'"'est fournie tel que permis selon la loi.
')
L_USAGE=(
'Usage: $prog [OPTIONS] [FILE(S)]...
  Blind test game : play extracts of audio or video FILE(S)
  The goal of the blind test game is to reconnize the extract.
  $prog uses mplayer to play extracts.
  
  Options:
  	-b : start playing at begin
             (the default is to start at a random position)
        -n <NUMBER> : stop after playing NUMBER files.
        -d <TIME> : number of second of the extracts (0 means dont stop)
                    default is 20
        -o <FILE> : play nothing but save the blindtest in the FILE
                    format depends on file extension :
                    .avi, .flv, .wmv for video
                    .wav, .mp3 (use lame), .ogg (use oggenc) for audio
                    the name of the extracts are written in .txt file
        -q <QUALITY> : only with -o option :
                       quality of the output file (1-9)
	-h : display this help
        -v : display version

  Usage sample:
	$ blindtest -n 20 /data/mp3/NightWish/
          this will play 20 files from directory /data/mp3/NightWish/
        $ blindtest -n 10 -d 10 -o toto.mp3 /data/mp3/NightWish/
          this will save a blindtest in toto.avi
'
'Usage: $prog [OPTIONS] [FICHIER(S)]...
  Jeu de blind test :
  Joue des extraits des FICHIER(S) audio ou video passes en argument.
  Le but du blind test est de reconnatre l'"'"'extrait.
  $prog utilise mplayer pour jouer les extraits.
  
  Options :
  	-b : commence a lire le fichier au dbut
             (par dfaut, la lecture commence  une position alatoire)
        -n <NOMBRE> : joue seulement NOMBRE fichiers et quitte
        -d <TEMPS> : nombre de seconde pour les extraits
                    par dfaut 20.
                    0 dsactive cette option.
        -o <FICHIER> : sauvegarde le rsultat du blind test dans FICHIER
                       le format de sortie dpend de l'"'"'extension:
                       .avi, .flv, .wmv pour les video
                       .wav, .mp3, .ogg pour l'"'"'audio
                       (lame/oggenc est requis pour le mp3/ogg)
                       le nom des extraits est enregistr dans un fichier .txt
        -q <QUALITE> : seulement avec l'"'"'option -o :
                       qualit du fichier de sortie (1-9)
	-h : affiche cette aide
        -v : affiche la version

  Exemples :
	$ blindtest -n 20 /data/mp3/NightWish/
          joue 20 fichiers pris dans l'"'"'arborescence /data/mp3/NightWish/
        $ blindtest -n 10 -d 10 -o toto.avi /data/mp3/NightWish/
          enregistre un blind test dans toto.avi
')

L_OPT=(
'Invalid option: $OPTARG. Try $prog -h'
'Option invalide: $OPTARG. Essayez $prog -h'
)
L_OPT_VALUE=(
'Invalid value for option $OPT. Try $prog -h'
'Valeur invalide pour l'"'"'option $OPT. Essayez $prog -h'
)
L_PLAY=(
'


          Playing audio/video file, press ENTER to stop, SPACE to pause'
'



                        Lecture du fichier audio/video
           Appuyez sur ENTER pour arrter, ESPACE pour mettre en pause'
)
L_ENTETE=(
'
                         ****************************
                         ***    Blind Test v$VERSION   ***
                         ****************************

'
'





                         ****************************
                         ***    Blind Test v$VERSION   ***
                         ****************************



')
L_MENU=(
'
     Titre : $TITRE

                    n  :  $RET
                    r  :  replay an extract
                    b  :  replay from begin
                    Q  :  quit
'
'

     Titre : $TITRE


                    n  :  $RET
                    r  :  rejoue un extrait
                    b  :  rejoue depuis le dbut
                    Q  :  quitter
')
L_DISPLAY=(
'display the song title'
'affiche le titre de la chanson'
)
L_NEXT=(
'goto next song'
'passe  la chanson suivante'
)
L_NEED_MPLAYER=(
'mplayer is not found. You must install it to.'
'Ce programme utilise mplayer. Veuillez l'"'"'installer pour continuer.'
)
L_NEED_LAME=(
'lame is not found. You must install it to encode to mp3.'
'Ce programme utilise lame. Veuillez l'"'"'installer pour continuer.'
)
L_NEED_OGGENC=(
'oggenc is not found. You must install it to encode to ogg.'
'Ce programme utilise oggenc. Veuillez l'"'"'installer pour continuer.'
)
L_UNKNOW_FORMAT=(
'Unknow or unsupported format: $OUTFILE
Supported file extension are: .avi, .mwv, .flv, .mp3, .ogg, .wav'
'Extension de fichier non support/Format inconnu: $OUTFILE
Les extensions valides sont : .avi, .mwv, .flv, .mp3, .ogg, .wav'
)
### End of locales


### Variables & Options

# TODO
# Entre des noms de joueur, compter les points, enregistrer...
# Enregistrer les extraits dans un fichier + fichier solution
# analyser type son et mettre en stereo/mono selon
# utiliser dialog, xdialog, kdialog, zenity ?


# duration of the extracts
declare -i DURATION=20
# start of the extracts
BEGIN=0
# output file
OUTFILE=
# maximum number of files
declare -i MAXFILE=0
# Quality for encoding
declare -i QUALITY=0

# The file(s) list
declare -i NB=0
declare -a LIST

# for i18n : display text in current language (use LC_ALL)
loc() {
    case $LC_ALL in
        fr*) echo "$(eval echo \""$2"\")" ;;
        *) echo "$(eval echo \""$1"\")" ;;
    esac
}

# display version and exit
version () {
cat << EOF
$prog $VERSION
$(loc "${L_VERSION[@]}")
EOF
exit
}

# display usage and exit
usage() {
    loc "${L_USAGE[@]}"
    exit 0
}

# Display an error message
error() {
    echo "$prog: $(loc "$@")" >&2
    exit 1
}

# Treat options
[ "$1" == "" ] && usage
[ "$1" == "--help" ] && usage
[ "$1" == "--version" ] && version

declare -i opts=0

while getopts ":hvbd:o:n:q:" i; do
  case "$i" in
    h*) usage ;;
    v*) version ;;
    d) DURATION=$OPTARG
          [ $DURATION = 0 ] && DURATION=32000
          opts=$((opts + 2)) ;;
    b) BEGIN=1; opts=$((opts + 1)) ;;
    o) OUTFILE="$OPTARG" ; opts=$((opts + 2)) ;;
    q) QUALITY=$OPTARG ; opts=$((opts + 2)) ;;
    n) MAXFILE=$OPTARG ; opts=$((opts + 2)) ;;
    *) error "${L_OPT[@]}" ;;
  esac
done

shift $opts

# output type
# Not that a quality is associed by default for each type
v=flv
case $(echo "$OUTFILE" | tr A-Z a-z) in
    *.flv) v=flv
        [ $QUALITY = 0 ] && QUALITY=7 ;;
    *.wmv) v=wmv2
        [ $QUALITY = 0 ] && QUALITY=5 ;;
    *.avi) v=mpeg4
        [ $QUALITY = 0 ] && QUALITY=3 ;;
    *.mp3) v=0
        which lame &>/dev/null || error "${L_NEED_LAME[@]}" ;
        [ $QUALITY = 0 ] && QUALITY=3 ;;
    *.ogg) v=0
        which oggenc &>/dev/null || error "${L_NEED_OGGENC[@]}" ;
        [ $QUALITY = 0 ] && QUALITY=3 ;;
    *.wav) v=0
        [ $QUALITY = 0 ] && QUALITY=5 ;;
#    *) [ $QUALITY = 0 ] && QUALITY=5 ;;
    "") [ $QUALITY = 0 ] && QUALITY=5 ;;
    *) error "${L_UNKNOW_FORMAT[@]}" ;;
esac

# Define video size for each quality
# Since the user do not want to set it
case $QUALITY in
    1)  ENC_1="-vf scale=640:480 -af resample=44100:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=2500 -lameopts mode=1:q=1" ;;
    2)  ENC_1="-vf scale=640:480 -af resample=44100:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=1500 -lameopts mode=1:q=2" ;;
    3)  ENC_1="-vf scale=512:384 -af resample=44100:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=1200 -lameopts mode=1:q=3" ;;
    4)  ENC_1="-vf scale=480:360 -af resample=22050:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=800 -lameopts mode=1:q=5" ;;
    5)  ENC_1="-vf scale=340:288 -af resample=22050:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=500 -lameopts mode=1:q=5" ;;
    6)  ENC_1="-vf scale=320:240 -af resample=22050:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=350 -lameopts mode=1:q=6" ;;
    7)  ENC_1="-vf scale=320:240 -af resample=22050:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=250 -lameopts mode=1:q=7" ;;
    8)  ENC_1="-vf scale=200:150 -af resample=22050:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=100 -lameopts mode=1:q=8" ;;
    9)  ENC_1="-vf scale=160:120 -af resample=8000:1:2"
        ENC_2="-lavcopts vcodec=$v:vbitrate=50 -lameopts mode=1:q=9:aq=9" ;;
    *) OPT='-q'; error "${L_OPT_VALUE[@]}" ;;
esac

# ENC_2="-lavcopts vcodec=$v:vbitrate=2000 -lameopts mode=1:q=1" ;;
# ENC_2="-ovc lavc -oac mp3lame $ENC_2"
ENC_1=" -af format=s16le -af pan=1:0.5:0.5 $ENC_1"
ENC_2="-ovc lavc -oac mp3lame $ENC_2"

# Remove options for command line
# TODO
continue=yes
while [ $continue = yes -a $# -gt 0 ] ; do
  case $1 in
    --) shift; continue=no ;;
    -*) shift;;
    *) continue=no ;;
  esac
done

# Test if mplayer is installed
which mplayer &>/dev/null || error "${L_NEED_MPLAYER[@]}" ;


### Program begin here

# Shuffle the files given as arguments and print the result
shuffle2 () {
    perl -e '
    srand;
    @new = ();
    while (@ARGV) {
        push(@new, splice(@ARGV, rand @ARGV, 1) . "\n");
    }
    print @new;
    ' "$@"
}

# [ -d toto ] && find ...

shuffle () {
    perl -e '
@a=();

sub fill {
    my $path = splice (@_, 0, 1);
    for (@_) {
        if ($_ ne ".." && $_ ne ".") {
            my $file = "$path$_";
            if ( -d "$file") {
                if (opendir(DIRHANDLE, "$file")) {
                    my @dir = readdir(DIRHANDLE);
                    splice (@dir, 0, 2);
                    fill ($file . "/", @dir);
                } else {
                    print "ERROR open: $file\n";
                    # TODO ERROR
                }
                closedir DIRHANDLE;
            } else {
                push(@a, "$file");
            }
        }
    }
}
fill "", @ARGV;

srand;
@new = ();
while (@a) {
    push(@new, splice(@a, rand @a, 1) . "\n");
}
print @new;

' "$@"
}


# Built the array of shuffle file name
make_shuffle_array () {
    TMP=$(mktemp)
    shuffle "$@" > $TMP
    while read a; do
        NB=$((NB+1))
        LIST[$NB]="$a"
        if [ ! "$MAXFILE" = 0 -a "$NB" -ge "$MAXFILE" ]; then
            break;
        fi
    done < $TMP
    rm $TMP
}

# Play a file
play () {
    clear
    loc "${L_ENTETE[@]}"
    loc "${L_PLAY[@]}"
    if [ "$2" = all ]; then
        mplayer -fs "$1" &>/dev/null
        return;
    fi
    RAND=$(get_begin_position "$1")
    #echo  mplayer -fs -ss $RAND -endpos $DURATION "$1"
    mplayer -fs -ss $RAND -endpos $DURATION "$1" &>/dev/null
}

get_begin_position () {
    local SIZE
    local RAND

    SIZE="$(mplayer -identify "$1" -vo null -nosound -frames 0 2>&1 |
          perl -ne 'print $1 if (/ID_LENGTH=([0-9]+)/)')"

    if [ ! "$SIZE" ]; then
        SIZE=0
    fi

    SIZE=$((SIZE-DURATION))
    if [ $SIZE -lt 1 ]; then
        SIZE=1
    fi

    RAND=$((RANDOM % SIZE))
    [ $BEGIN = 1 ] && RAND=0

    echo $RAND
}

# The blind function
blind () {
    FILE="$1"
    play "$FILE"
    RET="$(loc "${L_DISPLAY[@]}")"
    TITRE="???"

    CONTINUE=1
    while [ $CONTINUE = 1 ] ; do
    clear
    loc "${L_ENTETE[@]}"
    loc "${L_MENU[@]}"
        read -n 1 -s a
        case $a in
            r) play "$FILE" ;;
            b) play "$FILE" all ;;
            Q) exit 0 ;;
            n) if [ "$TITRE" = "???" ]; then
                   TITRE="$(basename "$FILE")"
                   if [ ! "$TITRE" = "$FILE" ]; then
                       TITRE="$TITRE"$'\n'" "${FILE%/*}""
                   fi
                   RET="$(loc "${L_NEXT[@]}")"
                else
                   CONTINUE=0
                fi ;;
        esac
    done
}

save_file () {
    local CMD=""
    local TXT="${OUTFILE%.*}.txt"
    : > $TXT
    while [ $NB -gt 0 ]; do
        FILE="${LIST[$((NB--))]}"
        echo "$FILE" >> $TXT
        POS=$(get_begin_position "$FILE")
        CMD="$CMD $(protect "$FILE") -ss $POS -endpos $DURATION $ENC_1 "
    done
    echo mencoder -o $(protect "$OUTFILE") $ENC_2 $CMD
    eval mencoder -o $(protect "$OUTFILE") $ENC_2 $CMD
}

save_mp3_file () {
    local TXT="${OUTFILE%.*}.txt"
    : > $TXT
    TMP=$(mktemp)
    rm $TMP
    TMP=$(echo $TMP | tr . _).wav
    while [ $NB -gt 0 ]; do
        FILE="${LIST[$((NB--))]}"
        echo "$FILE" >> $TXT
        POS=$(get_begin_position "$FILE")
        mplayer -ao pcm:file=/dev/stdout \
                -ss $POS -endpos $DURATION $ENC_1 "$FILE" >> $TMP$NB
    done
    sox $TMP* $TMP
    case "$OUTFILE" in
        *.mp3)
            lame -m m -q $QUALITY $TMP "$OUTFILE" ;;
        *.ogg)
            oggenc -q $QUALITY -o "$OUTFILE" $TMP ;;
        *.wav)
            mv $TMP "$OUTFILE" ;;
    esac
    rm $TMP*
}

protect () {
    echo -n "$1" | perl -pe 's:([^-a-zA-Z0-9/.]):\\$1:g;'
}

make_shuffle_array "$@"

if [ "$OUTFILE" ]; then
    if [ $v = 0 ]; then
        save_mp3_file
    else
        save_file
    fi
else
    while [ $NB -gt 0 ]; do
        blind "${LIST[$((NB--))]}"
    done
fi

exit 0
